#!/usr/bin/env python
from plotext._utility import all_markers, colors
import argparse, sys, os
import plotext as plt
try:
    import shtab
except ImportError:
    from . import _shtab as shtab

# For Possible Colors and Markers Completion 
def dict_to_complete(d={}):
    return {'zsh': '((' + ' '.join(map(lambda x: str(x[0]) + '\\:' + x[1], d.items())) + '))'}

def list_to_complete(data):
    return {'zsh': '((' + ' '.join([el + '\\:' for el in data]) + '))'}


def build_parser():
    
    examples = """Access each function documentation for further guidance, eg: plotext scatter -h"""

    parser = argparse.ArgumentParser(
        prog            = "plotext",
        description     = "plotting directly on terminal",
        epilog          = examples,
        formatter_class = argparse.RawDescriptionHelpFormatter)
    shtab.add_argument_to(parser,  ["-s", "--print-completion"])

    parser.set_defaults(type = "scatter")

    path_parser = argparse.ArgumentParser(add_help = False)

    path_parser.add_argument("-p", "--path",
                             action  = 'store',
                             #dest   = 'path',
                             type    = str,
                             metavar = "FILE",
                             help    = "file path of the data table; if not used it will read from stdin. Use 'test' to automatically download, in your user folder, some test data/image/gif or video, depending on the function used; the file will be removed after the plot",
                             ).complete = shtab.FILE
    path_parser.add_argument("-r", "--first-row",
                             action = "store",
                             type = int,
                             default = 0,
                             metavar = "FIRST-ROW",
                             help = "The first line to consider in the data file (counting from 0).")

    common_parser = argparse.ArgumentParser(add_help = False)

    common_parser.add_argument("-clt", "--clear_terminal",
                               nargs   = 1,
                               type    = str,
                               default = ["False"],
                               choices = ["True", "False"],
                               metavar = "BOOL",
                               help    = "it clears the terminal before plotting, if True (as by default)")

    common_parser.add_argument("-s", "--sleep",
                               nargs   = 1,
                               type    = float,
                               default = [0],
                               help    = "it adds a sleeping time after plotting, to reduce flickering when multiple plots are required; 0 by default")

    data_parser = argparse.ArgumentParser(add_help = False)

    data_parser.add_argument("-xcol", "--xcolumn",
                             nargs   = "+", # 1 or more
                             type    = str,
                             default = ["none"],
                             help    = "the column number (starting from 1), in the data table, that will be used as x data; by default 'none'")

    data_parser.add_argument("-ycols", "--ycolumns",
                             nargs   = "+", # 1 or more
                             type    = str,
                             default = ["all"],
                             help    = "the column numbers (starting from 1), in the data table, that will be used as y data; by default 'all'")

    data_parser.add_argument("-d", "--delimiter",
                             type    = str,
                             default = [' '],
                             nargs   = 1,
                             help    = "character to be used to separate columns in the data table (eg: ;); by default the white space ' '")

    data_parser.add_argument("-l", "--lines",
                             nargs   = "+", # 1 or more
                             type    = int,
                             default = [1000],
                             help    = "number of lines, from data table, to be plotted at each iteration; 1000 by default; data shorter then this will be plotted in a single iteration")

    options_parser = argparse.ArgumentParser(add_help = False)

    options_parser.add_argument("-m", "--marker",
                                type    = str,
                                default = ['hd'],
                                nargs   = 1,
                                choices = all_markers.keys(),
                                metavar = "marker",
                                help    = "character or marker code identifying the data points in the plot, eg: x, braille, sd, heart, fhd; by default hd",
                                ).complete = dict_to_complete(all_markers)
    
    options_parser.add_argument("-c", "--color",
                                type    = str,
                                default = [None],
                                nargs   = 1,
                                choices = colors,
                                metavar = "COLOR",
                                help    = "color of the data points in the plot, between: " + ",".join(colors) + "; add + at the end of the color (except black and white) for clearer colors, eg: red+"
                                ).complete = list_to_complete(colors)

    options_parser.add_argument("-t", "--title",
                                nargs   = 1,
                                type    = str,
                                default = [None],
                                help    = "the plot title")

    options_parser.add_argument("-xl", "--xlabel",
                                nargs   = 1,
                                type    = str,
                                default = [None],
                                help    = "the label of the x axis")

    options_parser.add_argument("-yl", "--ylabel",
                                nargs   = 1,
                                type    = str,
                                default = [None],
                                help    = "the label of the y axis")

    options_parser.add_argument("-g", "--grid",
                                nargs   = 1,
                                type    = str,
                                default = ["False"],
                                choices = ["True", "False"],
                                metavar = "BOOL",
                                help    = "it add grid lines if True, or removed them if False (as by default)")

    barhist_parser = argparse.ArgumentParser(add_help = False)

    barhist_parser.add_argument("-o", "--orientation",
                                nargs   = 1,
                                type = str,
                                default = ['v'],
                                choices = ['v', 'h'],
                                metavar = "ORIENTATION",
                                help    = "bar orientation, v for vertical (as by default), h for horizontal")

    barhist_parser.add_argument("-f", "--fill",
                                nargs   = 1,
                                type = str,
                                default = ["True"],
                                choices = ["True", "False"],
                                metavar = "BOOL",
                                help    = "it fills the bars with colored markers if True (as by default), otherwise only the bars skeleton is shown")

    subparser = parser.add_subparsers(dest = 'type',
                                      help = 'the user defined plot type')

    scatter = subparser.add_parser('scatter',
                                   description = "plots a series of data points",
                                   parents = [path_parser, data_parser, common_parser, options_parser],
                                   help = 'plots a series of data points',
                                   epilog = "eg: plotext scatter --path test --xcolumn 1 --ycolumns 2 --lines 5000 --title 'Scatter Plot Test' --marker braille")


    plot = subparser.add_parser('plot',
                                parents = [path_parser, data_parser, common_parser, options_parser],
                                description = "plots lines between consecutive data points",
                                help = 'plots lines between consecutive data points',
                                epilog = "eg: plotext plot --path test --xcolumn 1 --ycolumns 2 --sleep 0.1 --lines 2500 --clear_terminal True --color magenta+ --title 'Plot Test'")

    plotter = subparser.add_parser('plotter',
                                   parents = [path_parser, data_parser, common_parser, options_parser],
                                   description = 'plots a series of data points and the lines between consecutive ones',
                                   help = 'scatter + plot',
                                   epilog = "eg:plotext plotter --path test --xcolumn 1 --ycolumns 2 --sleep 0.1 --lines 120 --clear_terminal True --marker hd --title 'Plotter Test'")

    bar = subparser.add_parser('bar',
                               parents = [path_parser, data_parser, common_parser, options_parser, barhist_parser],
                               description = 'builds a bar plot',
                               help = 'bar plot',
                               epilog = "eg: plotext bar --path test --xcolumn 1 --title 'Bar Plot Test' --xlabel Animals --ylabel Count")

    bar.add_argument("-w", "--width",
                     nargs   = 1,
                     type    = float,
                     default = [None],
                     help    = "bars width as a float between 0 and 1")

    hist = subparser.add_parser('hist',
                                parents     = [path_parser, data_parser, common_parser, options_parser, barhist_parser],
                                description = 'builds a histogram plot',
                                help        = 'histogram plot',
                                epilog      = "eg: plotext hist --path test --xcolumn 1 --ycolumns 2 --lines 5000 --title 'Histogram Test'")

    hist.add_argument("-b", "--bins",
                      nargs   = 1,
                      type    = int,
                      default = [10],
                      help    = "histogram bins (10 by default)")

    image = subparser.add_parser('image',
                                 parents     = [path_parser, common_parser],
                                 description = 'plots an image from path',
                                 help        = 'plots an image from file path',
                                 epilog      = "eg: plotext image --path test")

    gif = subparser.add_parser('gif',
                               parents     = [path_parser, common_parser],
                               description = 'plays a gif image from path',
                               help        = 'plays a gif image from path',
                               epilog      = "eg: plotext gif --path test")

    video = subparser.add_parser('video',
                                 parents     = [path_parser, common_parser],
                                 description = 'plays a video from path',
                                 help        = 'plays a video from path',
                                 epilog      = "eg: plotext video --path test --from_youtube True")

    video.add_argument("-fy", "--from_youtube",
                       nargs   = 1,
                       type    = str,
                       default = ["False"],
                       choices = ["True", "False"],
                       metavar = "BOOL",
                       help    = "set it to True to render the colors correctly for videos downloaded from youtube; default is False")

    youtube = subparser.add_parser('youtube',
                                   description = 'plays a youtube video from url',
                                   help = 'plays a youtube video from url',
                                   epilog = "eg: plotext youtube --url test")

    youtube.add_argument("-u", "--url",
                         action  = 'store',
                         dest    = 'url',
                         nargs   = 1,
                         type    = str,
                         metavar = "URL",
                         help    = "the url of a youtube video; use 'test' for a test video")

    return parser


def main(argv = None):
    parser = build_parser()
    args = parser.parse_args(argv)

    type = args.type
    first_row = 0
    if type != "youtube":
        path = args.path
        first_row = args.first_row
        clt = True if args.clear_terminal[-1] == 'True' else False
        sleep = args.sleep[0]

    def get_xY(data):
        l = len(data)
        xcol = args.xcolumn[0]
        ycols = args.ycolumns
        xcol = "none" if xcol == "none" else int(xcol) if int(xcol) - 1 in range(l) else "none"
        all_ycols = [el for el in range(l) if el != xcol]
        ycols = all_ycols if ycols == ["all"] else [int(el) for el in ycols if int(el) - 1 in range(l)]
        x = list(range(1, len(data[0]) + 1)) if xcol == "none" else data[xcol - 1]
        Y = [data[i - 1] for i in ycols]
        return x, Y

    def plot(x, Y):
        for y in Y:
            plt.plot(x, y, marker = marker, color = color) if type in ['plot', 'plotter'] else None
            plt.scatter(x, y, marker = marker, color = None if type == 'plotter' else color) if type in ['scatter', 'plotter'] else None
            plt.bar(x, y, marker = marker, color = color, width = width, orientation = orientation, fill = fill) if type == 'bar' else None
            plt.hist(y, marker = marker, color = color, orientation = orientation, fill = fill, bins = bins) if type == 'hist' else None
        plt.clt() if clt else None
        plt.show()
        plt.cld()
        plt.sleep(sleep)

    data_plot = ['scatter', 'plot', 'plotter', 'bar', 'hist']
    test_path = 'test'

    if type in data_plot:
        lines = args.lines[0]
        delimiter = args.delimiter[0]
        title = args.title[0]
        xlabel = args.xlabel[0]
        ylabel = args.ylabel[0]
        grid = True if args.grid[-1] == 'True' else False

        orientation = args.orientation[0] if type in ['bar', 'hist']  else 'v'
        fill = args.fill[0] == 'True' if type in ['bar', 'hist'] else False
        width = args.width[0]  if type == 'bar' else 0
        bins = args.bins[0] if type in 'hist' else None

        marker = args.marker[0]
        color = args.color[0]

        plt.title(title);
        plt.xlabel(xlabel)
        plt.ylabel(ylabel)
        plt.grid(grid)

        if path == test_path:
            plt.plotsize(None, plt.th() - 3)
            if type != "bar":
                plt.download(plt.test_data_url, test_path, log = True)
            else:
                plt.download(plt.test_bar_data_url, test_path, log = True)
            path = test_path

        if path is None:
            def plot_text(text):
                data = plt._utility.read_lines(text, delimiter = delimiter)
                data = plt.transpose(data)
                x, Y = get_xY(data)
                plot(x, Y)

            for _ in range(first_row):
                sys.stdin.readline()
            text = []
            i = 0
            for line in iter(sys.stdin.readline, ''):
                text.append(line)
                i+=1;
                if len(text) == lines:
                    plot_text(text)
                    text = []
            if len(text) > 0: # this is when there is some residual data to be plotted, not lines long
                plot_text(text)
                text = []

        else:
            data = plt.read_data(path, delimiter = delimiter, first_row = first_row)
            data = plt.transpose(data)
            x, Y = get_xY(data)
            chunks = len(x) // lines + (1 if len(x) % lines else 0)
            for c in range(chunks):
                xc = x[c * lines: (c + 1) * lines]
                Yc = [y[c * lines: (c + 1) * lines] for y in Y]
                plot(xc, Yc)

    elif type == 'image':
        if path == test_path:
            plt.plotsize(None, plt.th() - 3)
            plt.download(plt.test_image_url, test_path, log = True)
            path = test_path
        plt.image_plot(path, fast = True)
        plt.clt() if clt else None
        plt.show()

    elif type == 'gif':
        if path == test_path:
            plt.plotsize(None, plt.th() - 3)
            plt.download(plt.test_gif_url, test_path, log = True)
            path = test_path
        plt.play_gif(path)

    elif type == 'video':
        if path == test_path:
            plt.plotsize(None, plt.th() - 3)
            plt.download(plt.test_video_url, test_path, log = True)
            path = test_path
        from_youtube = True if args.from_youtube[-1] == 'True' else False
        plt.play_video(path, from_youtube)

    elif type == 'youtube':
        url = args.url[-1]
        url = plt.test_youtube_url if url == test_path else url
        plt.play_youtube(url)

    if os.path.isfile(test_path):
        plt.delete_file(test_path, True)


if __name__ == "__main__":
    sys.exit(main())






# This file contains some plotext functions which are only available to the top main level and not to sub figures (which are written in _figure.py and _monitor.py). These are functions which requires some coding and would be too long to be added directly in _core.py

from plotext._utility import marker_codes, hd_symbols, sin
from plotext._figure import _figure_class
from plotext._utility import themes as _themes
import plotext._utility as ut
from time import time, sleep
from math import sqrt, ceil
import datetime as dt

figure = _figure_class() # the main figure at top level

##############################################
#######     Simple  Bar Functions    ########
##############################################

def simple_bar(*args, width = None, marker = None, color = None, title = None):
    x, y = ut.set_data(*args)
    marker = ut.correct_marker(marker)

    color_ok = ut.is_color(color) or (isinstance(color, list) and len(color) == len(x))
    color = [color] if color_ok else None

    simple_stacked_bar(x, [y], width = width, marker = marker, colors = color, title = title)

def simple_stacked_bar(*args, width = None, marker = None, colors = None, title = None, labels = None):
    x, y, Y, width = ut.bar_data(*args, width = width)
    marker = ut.correct_marker(marker)
    
    bars = len(Y); stacked_bars = len(Y[0])

    colors_ok1 = isinstance(colors, list) and isinstance(colors[0], list) and ut.matrix_size(colors) == [bars, stacked_bars]
    colors_ok2 = isinstance(colors, list) and len(colors) == stacked_bars
    colors = ut.transpose(colors) if colors_ok1 else [colors] * bars if colors_ok2 else [ut.color_sequence[:stacked_bars]] * bars

    title = ut.get_title(title, width)
    bars = [ut.single_bar(x[i], Y[i], y[i], marker, colors[i]) for i in range(bars)]
    labels = ut.get_simple_labels(marker, labels, colors[0], width)
    figure.monitor.matrix.canvas = title + '\n'.join(bars) + labels 
    figure.monitor.fast_plot = True

def simple_multiple_bar(*args, width = None, marker = None, colors = None, title = None, labels = None):
    x, y, Y, width = ut.bar_data(*args, width = width, mode='multiple')
    bars = len(Y); multiple_bars = len(Y[0]); lx = len(x[0])
    marker = ut.correct_marker(marker)

    colors_ok = isinstance(colors, list) and len(colors) == multiple_bars
    colors = colors if colors_ok else ut.color_sequence[:multiple_bars]

    out = ut.get_title(title, width)
    for i in range(bars):
        xn = [x[i] if j == (multiple_bars - 1) // 2 else ut.space * lx for j in range(multiple_bars)]
        new = [ut.single_bar(xn[j], [Y[i][j]], y[j][i], marker, [colors[j]]) for j in range(multiple_bars)]
        out += '\n'.join(new)
        out += '\n\n' if i != bars - 1 else ''
    labels = ut.get_simple_labels(marker, labels, colors, width)
    figure.monitor.matrix.canvas = out + labels
    figure.monitor.fast_plot = True

##############################################
#############     Play GIF    ################
##############################################

def play_gif(path):
    from PIL import Image, ImageSequence
    path = ut.correct_path(path)
    if not ut.is_file(path):
        return
    im = Image.open(path)
    index = 1
    for image in ImageSequence.Iterator(im):
        load_time = time()
        figure.clt()
        image = image.convert('RGB')
        figure.monitor._draw_image(image, fast = True)
        figure.show()
        load_time = time() - load_time
        frame_time = image.info['duration'] / 10 ** 3
        if load_time < frame_time:
            sleep(frame_time - load_time)
                
##############################################
##########     Video Functions    ############
##############################################

def play_video(path, from_youtube = False):
    path = ut.correct_path(path)
    if not ut.is_file(path):
        return
    _play_video(path, from_youtube)

def play_youtube(url):
    import pafy
    video = pafy.new(url)
    best = video.getbest()
    _play_video(best.url, from_youtube = True)

def get_youtube(url, path, log):
    import pafy
    video = pafy.new(url)
    best = video.getbest(preftype = "mp4")
    path = "youtube-video.mp4" if path is None else path
    path = ut.correct_path(path)
    best.download(filepath = path, quiet = not log)
    print(ut.format_strings('YouTube video downloaded as', path)) if log else None

def _play_video(path, from_youtube = False):
    import cv2
    from ffpyplayer.player import MediaPlayer
    from PIL import Image
    cap = cv2.VideoCapture(path)
    player = MediaPlayer(path)#, paused = True, loglevel = 'quiet');
    fr = 0;
    while fr == 0:
        fr = cap.get(cv2.CAP_PROP_FPS)
    frame_time = 1 / fr 
    #to_list = lambda frame: [[tuple(int(el) for el in tup) for tup in row] for row in frame]
    pt = lambda time: '{time:05.1f}  '.format(time=round(10 ** 3 * time, 1))
    real_time = video_time = 0
    while True:
        load_time = time()
        check_video, frame = cap.read();
        audio, check_audio = player.get_frame(show = False)
        load_time = time() - load_time
        if not check_video:
            break
        if load_time >= frame_time:
            continue
        real_time += load_time
        video_time += frame_time
        show_time = 0
        shown = False
        if video_time >= real_time:
            shown = True
            show_time = time()
            frame = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB) if from_youtube else frame
            #frame = to_list(frame)
            image = Image.fromarray(frame)
            figure.clt()
            figure.monitor._draw_image(image, fast = True)
            figure.show()
            show_time = time() - show_time
        sleep_time = 0
        if real_time < video_time:
            sleep_time = time()
            sleep(video_time - real_time)
            sleep_time = time() - sleep_time
        total_time = load_time + show_time + sleep_time
        real_time += show_time + sleep_time
        #print('load: ' + pt(load_time), 'show: ' + pt(show_time), 'sleep: ' + pt(sleep_time), 'total: ' + pt(total_time), 'frame: ' + pt(frame_time), 'real: ' + pt(real_time), 'video: ' + pt(video_time), 'r/v:', round(real_time / video_time, 3)) if shown else None
    player.close_player()
    cap.release()
    cv2.destroyAllWindows()
    figure.clf()

##############################################
############     Utilities     ###############
##############################################

test_data_url = "https://raw.githubusercontent.com/piccolomo/plotext/master/data/data.txt"
test_bar_data_url = "https://raw.githubusercontent.com/piccolomo/plotext/master/data/bar_data.txt"
test_image_url  = "https://raw.githubusercontent.com/piccolomo/plotext/master/data/cat.jpg"
test_gif_url    = "https://raw.githubusercontent.com/piccolomo/plotext/master/data/homer.gif"
test_video_url  = "https://raw.githubusercontent.com/piccolomo/plotext/master/data/moonwalk.mp4"
test_youtube_url = 'https://www.youtube.com/watch?v=ZNAvVVc4b3E&t=75s'

##############################################
#########     Matplotlib Backend    ##########
##############################################

def from_matplotlib(fig, marker = None):
    fig.canvas.draw()
    slots = (rows, cols) = fig.axes[0].get_subplotspec().get_gridspec().get_geometry()
    figure.clf(); #clt()
    figure.subplots(*slots)
    round10 = lambda data: [round(el, 10) for el in data]
    to_rgb = lambda rgb_norm: tuple([round(255 * el) for el in rgb_norm[:3]])
    figure.axes_color(to_rgb(fig.patch.get_facecolor()))
    for sub in fig.axes[:]:
        p = sub.get_subplotspec().get_geometry()[2]
        row = int((p - 0) / cols + 1)
        col = p + 1 - (row - 1) * cols
        monitor = figure.subplot(row, col)
        monitor.xlabel(sub.get_xlabel())
        monitor.ylabel(sub.get_ylabel())
        monitor.title(sub.get_title())
        monitor.xscale(sub.get_xscale())
        monitor.yscale(sub.get_yscale())
        monitor.xticks(round10(sub.get_xticks()))
        monitor.yticks(round10(sub.get_yticks()))
        monitor.canvas_color(to_rgb(sub.get_facecolor()))
        for point in sub.collections:
            label = point.get_label()
            label = label if label[0] != '_' else ''
            #point.set_offset_position('data')
            x, y = ut.transpose(point.get_offsets())
            color = [ut.to_rgb(point.to_rgba(el)) for el in point.get_facecolors()[0]]
            # can't find the right point colors
            monitor.scatter(x, y, label = label, marker = marker)
        for line in sub.get_lines():
            label = line.get_label()
            label = label if label[0] != '_' else ''
            x, y = line.get_data()
            monitor.plot(x, y, marker = marker, color = line.get_c(), label = label)
        for b in sub.patches:
            label = b.get_label()
            label = label if label[0] != '_' else ''
            color = b.get_facecolor()
            color = ut.to_rgb(color)
            box = b.get_bbox()
            x0, y0, x1, y1 = box.x0, box.y0, box.x1, box.y1
            x = [x0, x0, x1, x1, x0]
            y = [y0, y1, y1, y0, y0]
            fill = b.get_fill()
            fillx = fill if y0 == 0 else False
            filly = fill if x0 == 0 else False
            monitor.plot(x, y, fillx = fillx, filly = filly, marker = marker, color = color, label = label)
        monitor.xlim(*sub.get_xlim())
        monitor.ylim(*sub.get_ylim())

##############################################
#######     Presentation Functions    ########
##############################################
        
def markers():
    markers = list(hd_symbols.keys())[::-1] + list(marker_codes.keys())
    l = len(markers)
    rows = int(sqrt(l))
    cols = ceil(l / rows)
    y = ut.sin(1)
    figure.clf(); figure.theme('pro'); figure.xfrequency(0); figure.yfrequency(0); figure.frame(1)
    figure.subplots(rows, cols)
    figure.frame(0)
    for row in range(1, rows + 1):
        for col in range(1, cols + 1):
            i = (row - 1) * cols + col - 1
            if i < l:
                subplot = figure.subplot(row, col)
                figure.frame(1)
                default = ' [default]' if markers[i] == 'hd' else ''
                subplot.title(markers[i] + default)
                subplot.scatter(y, marker = markers[i])
                subplot.ticks_style('bold')
                #figure.ticks_color(figure._utility.title_color)
    figure.show()
    figure.clf()

def colors():
    print(ut.colorize("String Color Codes", style = 'bold'))
    bg = "default"
    c = ut.no_duplicates([el.replace('+', '') for el in ut.colors if el not in ['default', 'black', 'white']])
    cp = [ut.colorize(ut.pad_string(el + '+', 10), el + '+', background = bg) for el in c]
    c = [ut.colorize(ut.pad_string(el, 10), el, background = bg) for el in c]
    c = [' ' + c[i]  + cp[i] for i in range(len(c))]
    c = '\n'.join(c)
    print(' ' + ut.colorize(ut.pad_string('default', 20), background = bg))
    print(' ' + ut.colorize(ut.pad_string('black', 10), 'black', background = 'gray') + ut.colorize(ut.pad_string('white', 10), 'white', background = bg))
    print(c)
    print()
    #print(colorize("\n\nInteger Color Codes:", style = ''))
    c = ut.colorize("Integer Color Codes", style = 'bold', show = False) + '\n'
    for row in range(16):
        cr = ' '
        for col in range(16):
            i = row * 16 + col
            cr += ut.colorize(ut.pad_string(i, 5), i)
        c += cr + '\n'
    print(c)
    c = '\n'
    rgb = (100, 200, 85)
    rgb_string = '(' + ', '.join([str(el) for el in rgb]) + ')'
    print(ut.colorize("RGB Tuples like:", style = "bold"), ut.colorize(rgb_string, rgb, "bold"))

def styles():
    from plotext._utility import styles, colorize, title_color
    c = [colorize(el, style = el) for el in styles]
    c = '\n'.join(c)
    print(c)
    mul = 'bold italic dim'
    print('\n' + colorize('multiple styles are accepted, ', title_color) + 'eg: ' + colorize(mul, style = mul))

def themes():
    themes = list(_themes.keys())[::]
    l = len(themes)
    rows = int(sqrt(l))
    cols = ceil(l / rows)
    y1 = ut.sin(periods = 1)
    y2 = ut.sin(periods = 1, phase = -1)
    figure.clf()
    figure.subplots(rows, cols)
    for row in range(1, rows + 1):
        for col in range(1, cols + 1):
            i = (row - 1) * cols + col - 1
            if i < l:
                subplot = figure.subplot(row, col)
                subplot.theme(themes[i])
                subplot.title(themes[i])
                subplot.scatter(y1); subplot.plot(y2)
    figure.show()
    figure.clf()

##############################################
###########     Test Function    #############
##############################################

def test():
    import random
    figure.clf(); figure.clt()
    figure.date_form("d/m/Y");
    figure.take_min()
    figure.plot_size(None, ut.terminal_height())
    #figure.plot_size(108, 70)
    
    figure.plotsize(ut.tw(), ut.th() - 5)
    figure.subplots(2, 2)
    
    subplot = figure.subplot(1, 1)
    subplot.title("Multiple Axes Plot")
    subplot.canvas_color(66); subplot.axes_color(4); subplot.ticks_color(216); subplot.ticks_style('bold italic')
    y = ut.sin(periods = 1); l = len(y)
    subplot.scatter(y, label = "lower left")
    x = [figure.today_datetime() + dt.timedelta(days = i) for i in range(l)]; x = figure.datetimes_to_strings(x)
    subplot.plot(x, x, label = 'upper right - all dates', xside = 2, yside = 2)
    subplot.vline(l / 2, 'red')
    subplot.hline(0, 200)
    subplot.text("origin", l // 2, 0, color = 'red', alignment = 'center')
    subplot.xlabel('x lower'); subplot.xlabel('x upper', 2)
    subplot.ylabel('y left', 'left'); subplot.ylabel('y right', 'right')
    subplot.xfrequency(8); subplot.xfrequency(5, 2);
    subplot.yfrequency(3); subplot.yfrequency(5, 2);
    subplot.grid(1,1)
    
    subplot = figure.subplot(1, 2)
    subplot.theme('innocent')
    xb = ["Sausage", "Pepperoni", "Mushrooms", "Cheese", "Chicken", "Beef"]
    y1 = [36, 14, 11, 8, 7, 4]
    y2 = [20, 12, 35, 15, 4, 5]
    subplot.stacked_bar(xb, [y1, y2], labels = ["men", "women"])

    subplot = figure.subplot(2, 1)
    subplot.theme('dreamland')
    ld = 7 * 10 ** 4
    data = [random.gauss(0, 1) for el in range(10 * ld)]
    subplot.hist(data, bins = 60, label="mean 0")
    subplot.frame(1); #subplot.xaxes(1, 0); subplot.yaxes(1, 0)
    
    subplot = figure.subplot(2, 2)
    subplot.canvas_color('gray+'); subplot.axes_color('gray+')
    ut.download(test_image_url, 'cat.jpg')
    subplot.image_plot('cat.jpg', grayscale = False)
    ut.delete_file('cat.jpg')
    subplot.title('A very Cute Cat')
    subplot.frame(0)

    #figure.plotsize(0, 0)
    figure.show()
    figure._get_time()
    figure.save_fig('test.txt')
    figure.save_fig('test.html')
    #figure.clf()

